function GroupStrainer(fname)
    % parameters of LSH, corresponding to similarity thresholds
    r = [27 27 25 18 15 12 10 8];
    b = [6 26 86 68 86 78 76 96];
    % similarity threshold
    thresh = [0.9 0.85 0.8 0.75 0.7 0.65 0.6 0.55];
    data = textread(strcat('../input/',fname),'','delimiter',',');
    userMax = max(data(:,1))+1;
    prodMax = max(data(:,2))+1;
    A = sparse(data(:,2)+1,data(:,1)+1,1,prodMax,userMax);
    iter = 1;
    
    disp('Iteration:1')
    
    [k,l,Nx,Ny,Qy,DnZ] = JacClus(r(iter), b(iter), A);
    fid = fopen('../log/clusterLabels', 'w');  
    fprintf(fid, '%d ', Qy); 
    fclose(fid);
    python('../src/SPClusteringJac.py',num2str(thresh(iter)),fname);
    Qy = textread('../log/Qy','');
    Ny = textread('../log/Ny','');
    Qy = Qy';
    Ny = Ny';
    
    [k,l,Ny,DnZ] = compute_Dnz(Qy, Nx,Ny, A);
    
    fid = fopen(strcat('../log/clusterLabelsIter',num2str(iter)), 'w');  
    fprintf(fid, '%d ', Qy); 
    fclose(fid);
    
    iter = iter + 1;
    while(iter<9)
        disp(strcat('Iteration:',num2str(iter)))
        [Ny, Qy,cluster_ids] = CosLSH(r(iter),b(iter), k,l,Nx,Ny, Qy,DnZ);
        [l, Ny, DnZ] = compute_Dnz2(Ny,DnZ,cluster_ids);
        fid = fopen('../log/clusterLabels', 'w');  
        fprintf(fid, '%d ', Qy); 
        fclose(fid);
        python('../src/SPClusteringCos.py',num2str(thresh(iter)),num2str(iter),fname);
        Qy = textread('../log/Qy','');
        Ny = textread('../log/Ny','');
        Qy = Qy';
        Ny = Ny';
        [k,l,Ny,DnZ] = compute_Dnz(Qy,Nx,Ny,A);
        fid = fopen(strcat('../log/clusterLabelsIter',num2str(iter)), 'w');  
        fprintf(fid, '%d ', Qy); 
        fclose(fid);
        iter=iter+1;
    end
    fid = fopen('../output/UserLabels', 'w');  
    fprintf(fid, '%d\n', Qy); 
    fclose(fid);
exit
end

function [k,l,Nx,Ny,Qy,Dnz] = JacClus(r, b, A)
    [m,n] = size(A);
    Nx = ones(1,m);
    Ny = ones(1,n);
    Qy = (1:n);
    [Ny, Qy] = JacLSH(A, r, b, Ny, Qy);
    [k,l,Ny,Dnz] = compute_Dnz(Qy,Nx, Ny, A);
end

function [k,l,Ny,DnZ] = compute_Dnz(Qy, Nx,Ny, DnZ)
    uy = unique(Qy);
    Ny = Ny(uy);
    k = length(Nx);
    l = length(Ny);
    
    for j=1:l
        if(Ny(j)==1), continue;  end
        ind=find(Qy==uy(j));
        Qy(ind)=-1; Qy(ind(1))=uy(j);
        DnZ(:,ind(1)) = sum(DnZ(:, ind),2);
    end
    DnZ = DnZ(:,Qy>0);
end

function [Ny, Qy,cluster_ids] = CosLSH(r, b, k0,l0,Nx,Ny, Qy,DnZ)
    Dnz=DnZ;
    Nxy = Nx' * Ny;
    Pnz = Dnz./Nxy;
    s=r*b;
    S = zeros(s,l0);
    for i=1:s
        ri = randn(1,k0);
        ri = ri/sqrt(sum(ri.^2));
        S(i,:) = sign(ri*Pnz);
    end
    maps = cell(b,1);
    for j = 1:b
        from = 1 + r*(j-1);
        to = from + (r-1);
        c=containers.Map;
        for i = 1:l0   
            t = S(from:to,i)';   
            t = num2str(t); 
            if(isKey(c,t))
                c(t) = [c(t) i];
            else
                c(t) = i;
            end
        end
        maps{j} = c;
    end 
    clear S;
    candidategroups = 1:l0;
    for i=1:b
        c = maps{i};
        k=c.keys;
        for j=1:length(k)
           candidategroups(c(k{j})) = min(candidategroups(c(k{j}))); 
        end
    end
    ucg=unique(candidategroups);
    cluster_ids = unique(Qy);
    for i=1:length(ucg)
        ind = find(candidategroups == ucg(i));
        num=length(ind);
        cand1 = ind(1);
        for o=2:num
           cand1 = [cand1 ind(o)];
        end
        if(length(cand1)>1)
            newind = cand1;
            cis = Ny(newind);
            [minid minind] = min(cluster_ids(newind));
            Ny(newind) = 0;
            Ny(newind(minind)) = sum(cis);
            for q=1:length(newind)            
               if(cluster_ids(newind(q))~=minid)                
                    Qy(Qy==cluster_ids(newind(q))) = minid; 
                    cluster_ids(cluster_ids==cluster_ids(newind(q))) = minid;
               end
            end
        end
    end
end

function [l, Ny, DnZ] = compute_Dnz2(Ny,DnZ,cluster_ids)
    uy = unique(cluster_ids);
    Ny = Ny(Ny>0);
    l = length(Ny);
    for j=1:l
        if(Ny(j)==1), continue;  end
        ind=find(cluster_ids==uy(j));
        cluster_ids(ind)=-1; cluster_ids(ind(1))=uy(j);
        DnZ(:,ind(1)) = sum(DnZ(:, ind),2);
    end
    DnZ = DnZ(:,cluster_ids>0);
end

function [Ny, Qy] = JacLSH(A, r, b, Ny, Qy)
    [n,m] = size(A);
    s=r*b;
    M = zeros(s,n);
    for i=1:s
        M(i,:) = randperm(n);    
    end
    S = zeros(s,m);
    for i=1:m
        if(sum(A(:,i)>0)>0)
            S(:,i) = min(M(:,(A(:,i)>0)),[],2);    
        end
    end
    clear M;
    maps = cell(b,1);

    for j = 1:b
        from = 1 + r*(j-1);
        to = from + (r-1);
        c=containers.Map;
        for i = 1:m   
            t = S(from:to,i)';   
            t = num2str(t);  

            if(isKey(c,t))
                c(t) = [c(t) i];
            else
                c(t) = i;
            end
        end
        maps{j} = c;
    end
    clear S;
    candidategroups = 1:m;

    for i=1:b
        c = maps{i};
        k=c.keys;
        for j=1:length(k)
           candidategroups(c(k{j})) = min(candidategroups(c(k{j}))); 
        end
    end
    ucg=unique(candidategroups);
    for i=1:length(ucg)
        ind = find(candidategroups == ucg(i));
        num=length(ind);
        cand1 = ind(1);
        for o=2:num
            cand1 = [cand1 ind(o)];
        end

        if(length(cand1)>1)
           newind = cand1;
           Ny(newind) = 0;
           Ny(min(newind)) = length(newind);
           Qy(newind) = min(newind);
        end
    end % candidate groups
end