# this file is to to compare the ground truth with output of CoClusterLSH
import matplotlib.pyplot as plt
import numpy as np    
import math
import sys

def printMatrix(newEdges, matrixPath, userN, prodN, LSHClusters=[]):
    revUser = []
    revProd = []
    for edge in newEdges:
        revUser.append(edge[0])
        revProd.append(edge[1])
                                                                              
    fig, ax = plt.subplots(1, 1)
    fig.set_size_inches(20, 5)
    plt.xlabel('User ID')
    plt.ylabel('Prod ID')          
    major_ticks = np.arange(0, userN, 20)                                              
    minor_ticks = np.arange(0, userN, 5)                                               
    
    ax.set_xticks(major_ticks)                                                       
    ax.set_xticks(minor_ticks, minor=True)                                           
    ax.set_yticks(major_ticks)                                                       
    ax.set_yticks(minor_ticks, minor=True)    

    plt.xlim([-1, userN])
    plt.ylim([-1, prodN])
        
    plt.grid(True, which='both')
    plt.legend()
#     plt.title(title + " total #:" + str(len(values1)))
    plt.scatter(revUser, revProd, 3, c='r', marker=u'x')
    
    bins = []
    values = []
    if len(LSHClusters) != 0:
        sum = -1
        for cluster in LSHClusters:
            sum = sum + len(cluster)
            if len(cluster) > 10:
                bins.append(sum)
                values.append(prodN)
        plt.bar(bins, values, width=0.5)
    
    plt.savefig(matrixPath)

def readEdges(edgepath):
    edges = []
    userN = 0
    prodN = 0
    
    edgefile = open(edgepath)
    for line in edgefile:
        line = line.strip()
        units = line.split(',')
        user = int(units[0])
        prod = int(units[1])
        if user > userN:
            userN = user
        if prod > prodN:
            prodN = prod
        edges.append([user, prod])
    edgefile.close()
    userN = userN + 1
    prodN = prodN + 1
    matrix = [[0 for j in range(0, prodN)] for i in range(0, userN)]  # user as rows!!! 
    for edge in edges:
        matrix[edge[0]][edge[1]] = 1
        
    return matrix, userN, prodN,edges

def readLSHOut(inpath):
    groupDic = {}
    infile = open(inpath)
    for line in infile:
        line = line.strip()
        units = line.split(' ')
        userId = 0
        for unit in units:
            if not groupDic.has_key(int(unit)):
                groupDic[int(unit)] = set()
            groupDic[int(unit)].add(userId)
            userId = userId + 1
    infile.close()
    return groupDic

def cosineSim(vector1, vector2):
    numerator = 0
    denominator1 = 0
    denominator2 = 0
    for i in range(0, len(vector1)):
        numerator = numerator + vector1[i] * vector2[i]
        denominator1 = denominator1 + vector1[i] * vector1[i]
        denominator2 = denominator2 + vector2[i] * vector2[i]
        
    return numerator * 1.0 / (math.sqrt(denominator1) * math.sqrt(denominator2))
    
def SinglePassClustering(LSHGroupDicBef, LSHGroupDicAft, simThresh, edgesPath):
    matrix, _, _, edges = findSimThreshold.readEdges(edgesPath)
    groupsBef = LSHGroupDicBef.values()
    groupsAft = LSHGroupDicAft.values()
    newGroups = []

    # where the groups are in last iteration
    groupMapping = [[] for a in groupsAft] 
    for iBef in range(0, len(groupsBef)):
        # choose one of the members
        for user in groupsBef[iBef]:
            break
        for iAft in range(0, len(groupsAft)):
            if user in groupsAft[iAft]:
                groupMapping[iAft].append(iBef)
                break
    
    # check every thing is as expected, pure merge
    for iAft in range(0, len(groupsAft)):
        l1 = len(groupsAft[iAft])
        l2 = 0
        for ele in groupMapping[iAft]:
            l2 = l2 + len(groupsBef[ele])
        assert l1 == l2
    
    # compute group vector for cosine similarity
    groupVecs = []
    for userGroup in groupsBef:
        groupVec = []
        for user in userGroup:
            userVec = matrix[user]
            groupVec.append(userVec)
        vect = [0 for i in range(0, len(groupVec[0]))]
        for i in range(0, len(groupVec[0])):
            for j in range(0, len(groupVec)):
                vect[i] = vect[i] + groupVec[j][i]
            vect[i] = vect[i] * 1.0 / len(groupVec)
        groupVecs.append(vect)
        
    # single pass filtering
    filteredGroups = []
    for iAft in range(0, len(groupMapping)):
        clusterL = groupMapping[iAft]  # group ids in Bef
        newGroups = []
        for iBef in range(0, len(clusterL)):
            iBefClus = clusterL[iBef]
            flag2 = False
            for group in newGroups:
                flag1 = False
                for member in group:
                    sim = cosineSim(groupVecs[iBefClus], groupVecs[member])
                    if sim < simThresh:
                        flag1 = True
                        break
                if flag1 == False:
                    group.append(iBefClus)
                    flag2 = True
                    break
            if flag2 == False:
                newGroups.append([iBefClus])
        filteredGroups.extend(newGroups)
    
    # merge groups in filtered groups
    outputGroups = []
    for i in range(0, len(filteredGroups)):
        group = filteredGroups[i]
        userS = set()
        for member in group:
            userS = userS.union(groupsBef[member])
        outputGroups.append(userS)
    return outputGroups, matrix, edges

def changeUserId4plot(LSHClusters, edges):
    LSHClusters = sorted(LSHClusters, key=lambda cluster:len(cluster), reverse=True)
    count = 0
    idMap = {}
    for cluster in LSHClusters:
        for oldId in cluster:
            idMap[oldId] = count
            count = count + 1
    newEdges = []
    for edge in edges:
        user = idMap[edge[0]]
        prod = edge[1]
        newEdges.append([user, prod])
    return idMap, newEdges, LSHClusters

def outputQyNy(LSHClusters, QyPath, NyPath):
    idClusterDic = {}
    for clusIndex in range(0, len(LSHClusters)):
        cluster = LSHClusters[clusIndex]
        for ele in cluster:
            idClusterDic[ele] = clusIndex + 1
    sortedDicL = sorted(idClusterDic.items(), key=lambda x:x[0])
    Qy = [a[1] for a in sortedDicL]
    Ny = [len(a) for a in LSHClusters]
    qfile = open(QyPath, 'w')
    for q in Qy:
        qfile.write(str(q) + '\n')
    qfile.close()
    nfile = open(NyPath, 'w')
    for n in Ny:
        nfile.write(str(n) + '\n')
    nfile.close()

if __name__ == '__main__':
    noiseRate = ""
    simThresh = float(sys.argv[1]) 
    iter = int(sys.argv[2])
    inputName = sys.argv[3]
    
    LSHpath = '../log/'
    clusBefPath = LSHpath + 'clusterLabelsIter' + str(iter - 1)
    clusAftPath = LSHpath + 'clusterLabels'
    matrixPath = '../output/' +'outputMatrixVisualization' + '.pdf'
    edgesPath = '../input/' + inputName
    QyPath = LSHpath + 'Qy' + str(noiseRate)
    NyPath = LSHpath + 'Ny' + str(noiseRate)

    LSHGroupDicBef = readLSHOut(clusBefPath)
    LSHGroupDicAft = readLSHOut(clusAftPath)
    LSHClusters, matrix, edges = SinglePassClustering(LSHGroupDicBef, LSHGroupDicAft, simThresh, edgesPath)
    
    clusterSize = [len(ele) for ele in LSHClusters]
    userIdMap, newEdges, LSHClustersNew = changeUserId4plot(LSHClusters, edges)
    userN = len(matrix)
    prodN = len(matrix[0])
    printMatrix.printMatrix(newEdges, matrixPath, userN, prodN, LSHClustersNew)
    outputQyNy(LSHClusters, QyPath, NyPath)
