# this file is to to compare the ground truth with output of CoClusterLSH
import matplotlib.pyplot as plt
from numpy import histogram
import math
import sys

def readEdges(edgepath):
    edges = []
    userN = 0
    prodN = 0
    
    edgefile = open(edgepath)
    for line in edgefile:
        line = line.strip()
        units = line.split(',')
        user = int(units[0])
        prod = int(units[1])
        if user > userN:
            userN = user
        if prod > prodN:
            prodN = prod
        edges.append([user, prod])
    edgefile.close()
    userN = userN + 1
    prodN = prodN + 1
    matrix = [[0 for j in range(0, prodN)] for i in range(0, userN)]  # user as rows!!! 
    for edge in edges:
        matrix[edge[0]][edge[1]] = 1
        
    return matrix, userN, prodN,edges
    
def readLSHOut(inpath):
    groupDic = {}
    infile = open(inpath)
    for line in infile:
        line = line.strip()
        units = line.split(' ')
        userId = 0
        for unit in units:
            if not groupDic.has_key(int(unit)):
                groupDic[int(unit)] = set()
            groupDic[int(unit)].add(userId)
            userId = userId + 1
    infile.close()
    return groupDic

def JaccardSim(vector1, vector2):
    andCount = 0
    orCount = 0
    for i in range(0, len(vector1)):
        if vector1[i] + vector2[i] == 2:
            andCount = andCount + 1
        if vector1[i] + vector2[i] >= 1:
            orCount = orCount + 1
    return andCount * 1.0 / orCount

def SinglePassClustering(LSHGroupDic, simThresh, edgesPath):
    matrix, _, _, edges = readEdges(edgesPath)
    print len(matrix)
    newGroups = []
    for key in LSHGroupDic:
        group = LSHGroupDic[key]
        tempNewGroups = []
        for user in group:
            flag1 = False
            for newGroup in tempNewGroups:
                flag2 = False
                for tempUser in newGroup:
                    if JaccardSim(matrix[user], matrix[tempUser]) < simThresh:
                        flag2 = True
                        break
                if flag2 == False:
                    newGroup.add(user)
                    flag1 = True
                    break
            if not flag1:
                tempNewGroups.append(set([user]))
        newGroups.extend(tempNewGroups)
    return newGroups, matrix, edges

def changeUserId4plot(LSHClusters, edges):
    LSHClusters = sorted(LSHClusters, key=lambda cluster:len(cluster), reverse=True)
    count = 0
    idMap = {}
    for cluster in LSHClusters:
        for oldId in cluster:
            idMap[oldId] = count
            count = count + 1
    newEdges = []
    for edge in edges:
        user = idMap[edge[0]]
        prod = edge[1]
        newEdges.append([user, prod])
    return idMap, newEdges, LSHClusters

def outputQyNy(LSHClusters, QyPath, NyPath):
    idClusterDic = {}
    for clusIndex in range(0, len(LSHClusters)):
        cluster = LSHClusters[clusIndex]
        for ele in cluster:
            idClusterDic[ele] = clusIndex + 1
    sortedDicL = sorted(idClusterDic.items(),key = lambda x:x[0])
    Qy = [a[1] for a in sortedDicL]
    Ny = [len(a) for a in LSHClusters]
    qfile = open(QyPath,'w')
    for q in Qy:
        qfile.write(str(q)+'\n')
    qfile.close()
    nfile = open(NyPath,'w')
    for n in Ny:
        nfile.write(str(n)+'\n')
    nfile.close()

if __name__ == '__main__':
    noiseRate = ""
    simThresh = float(sys.argv[1])
    inputName = sys.argv[2]
    
    LSHpath = '../log/'
    LSHOutPath = LSHpath + 'clusterLabels' + str(noiseRate)
    edgesPath = '../input/' + inputName
    QyPath = LSHpath + 'Qy' + str(noiseRate)
    NyPath = LSHpath + 'Ny' + str(noiseRate)

    LSHGroupDic = readLSHOut(LSHOutPath)
    LSHClusters = LSHGroupDic.values()
    LSHClusters, matrix, edges = SinglePassClustering(LSHGroupDic, simThresh, edgesPath)
    clusterSize = [len(ele) for ele in LSHClusters]
    userIdMap, newEdges, LSHClustersNew = changeUserId4plot(LSHClusters, edges)
    userN = len(matrix)
    prodN = len(matrix[0])
    outputQyNy(LSHClusters, QyPath, NyPath)
