from numpy import histogram
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
import matplotlib.cm as cm

from math import ceil, pow, log, floor, sqrt
import sys

class Config:
    highDegThresh = 20
    epslon = 1
    aDeg = 0.9
    bDeg = 3
    aPR = 0.1
    bPR = 0.3
    PRPath = ""
    outputDir = "" 
    edgeListPath = ""
    priorPath = ""
    figDir = ""
    
    def __init__(self, inputfileDir,prfileDir,outputDir,figDir):
        self.PRPath = prfileDir
        self.outputDir = outputDir
        self.edgeListPath = inputfileDir
        self.figDir = figDir
        self.priorPath = outputDir + "_Prod_NFS"

def readPR(PRFilePath):
    nodesPR = []
    count = 0
    inFile = open(PRFilePath)
    for line in inFile:
        count += 1
        nodesPR.append(float(line.strip()))
    inFile.close()
    return count, nodesPR

def getAdj(nodesPR, totalNodes, edgeListPath):
    prodAdj = {}
    userAdj = {}
    productIDs = set()
    inFile = open(edgeListPath)
    for line in inFile:
        edge = line.strip()
        nodes = edge.split(",")
        if len(nodes) != 2:
            continue
        userID = int(nodes[0])
        prodID = int(nodes[1])
        productIDs.add(prodID)
        if not prodAdj.has_key(prodID):
            prodAdj[prodID] = set()
        prodAdj[prodID].add(userID)
        if not userAdj.has_key(userID):
            userAdj[userID] = set()
        userAdj[userID].add(prodID)
    inFile.close()
        
    return prodAdj, userAdj,productIDs

def findHighDegProd(prodAdj, userAdj, nodesPR, neighborNumThresh):
    neighUserPRDistri = []
    neighUserDegDistri = []
    for prodID in prodAdj:
        neighUser = prodAdj[prodID]
        if len(neighUser) >= neighborNumThresh:
            neighUserDeg = []
            neighUserPR = []
            for userID in neighUser:
                neighUserDeg.append(len(userAdj[userID]))
                neighUserPR.append(nodesPR[userID])
            neighUserDegDistri.append([prodID, neighUserDeg])
            neighUserPRDistri.append([prodID, neighUserPR])
    return neighUserPRDistri, neighUserDegDistri

def vector2ProbHistPR(data, configure):
    a = configure.aPR
    b = configure.bPR
    epslon = configure.epslon
    xmax = max(data)
    xmin = min(data)
    bins = []
    minK = int(floor(log(xmax / a, b)))  # b is less than 1, thus mim of K correspond to xmax
    maxK = int(ceil(log(xmin / a, b)))
    for k in range(minK, maxK + 1):
        bins.append(a * pow(b, k))
    bins.reverse()
    hist, _ = histogram(data, bins)
    assert sum(hist) == len(data)
    histCopy = [value for value in hist]
    for i in range(0, len(histCopy)):
        if histCopy[i] == 0:
            histCopy[i] = epslon
    distri = [i * 1.0 / sum(histCopy) for i in histCopy]
    return distri, bins

def vector2ProbHistDeg(data,configure):
    a = configure.aDeg
    b = configure.bDeg
    epslon = configure.epslon
    xmax = max(data)
    xmin = min(data)
    bins = []
    minK = int(floor(log(xmin / a, b)))
    maxK = int(ceil(log(xmax / a, b)))
    for k in range(minK, maxK + 1):
        bins.append(a * pow(b, k))
    hist, _ = histogram(data, bins)
    assert sum(hist) == len(data)
    histCopy = [value for value in hist]
    for i in range(0, len(histCopy)):
        if histCopy[i] == 0:
            histCopy[i] = epslon
    distri = [i * 1.0 / sum(histCopy) for i in histCopy]
    return distri, bins

def computeKL(product, histAllPR, bins, epslon):
    neighbors = product[1]
    neighborsHist, _ = histogram(neighbors, bins)
    copyNeighborHist = [float(i) for i in neighborsHist]
    assert len(histAllPR) == len(copyNeighborHist)
    assert sum(copyNeighborHist) == len(neighbors)
    for i in range(0, len(copyNeighborHist)):
        if copyNeighborHist[i] == 0:
            copyNeighborHist[i] = epslon
    histSum = sum(copyNeighborHist)
    distriPR = [i / histSum for i in copyNeighborHist]
    KL = 0
    for i in range(0, len(histAllPR)):
        assert distriPR[i] != 0
        KL = KL + histAllPR[i] * log(histAllPR[i] / distriPR[i])
    return product[0], KL

def vector2FreqHistGivenBins(data, bins):
    hist, _ = histogram(data, bins)
    histCopy = [value for value in hist]
    return histCopy, bins

def computeEntro(product, bins):
    data = product[1]
    hist, _ = vector2FreqHistGivenBins(data, bins)
    assert len(data) == sum(hist)
    vecLen = len(data)
    entro = 0.0
    for i in range(0, len(hist)):
        if hist[i] > 0:
            prob = hist[i] * 1.0 / vecLen
            entro = entro - prob * log(prob)
    return product[0], entro

def plotPairHotColorMap(values1, label1, values2, label2, outputPath, title, intevals):
    
    value1Max = max([a[1] for a in values1])
    value1Min = min([a[1] for a in values1])
    value2Max = max([a[1] for a in values2])
    value2Min = min([a[1] for a in values2])
    dim1Inteval = (value1Max - value1Min) / intevals
    dim2Inteval = (value2Max - value2Min) / intevals
    normalPos = [[0 for i in range(0, intevals + 1)] for j in range(0, intevals + 1)]
    for i in range(0,len(values1)):
        index1 = int((values1[i][1] - value1Min) / dim1Inteval)
        index2 = int((values2[i][1] - value2Min) / dim2Inteval)
        normalPos[index2][index1] = normalPos[index2][index1] + 1
    
    _, ax = plt.subplots(1, 1)
    plt.xlabel(label1)
    plt.ylabel(label2)
    plt.grid(True)
    plt.legend()
    im = plt.imshow(normalPos, extent=[value1Min, value1Max, value2Min, value2Max], origin='lower', alpha=1, aspect='auto', cmap=plt.cm.cmap_d['hot'], norm=LogNorm(), interpolation=None)
    plt.colorbar()
    plt.title("High Degree "+title + " total #:" + str(len(values1)))
    plt.savefig(outputPath)

def plotFeatures(configure, prodPRKL, prodPREntro, prodDegKL, prodDegEntro):  
    plotPairHotColorMap(prodPRKL, "PR KL Divergence", prodPREntro, "PR Entropy", configure.figDir + "_prod_PR_KL_Entro.pdf", "products ", 300)
    plotPairHotColorMap(prodDegKL, "Degree KL Divergence", prodDegEntro, "Degree Entropy", configure.figDir + "_prod_Deg_KL_Entro.pdf", "products", 300)
    
def findRankRatio(values, rev=False):
    dic = {}
    newlist = []
    if rev:
        newlist = sorted(values, reverse=True)
    else:
        newlist = sorted(values)
    for index in range(0, len(newlist)):
        if dic.has_key(newlist[index]) == False:
            dic[newlist[index]] = 1 - index * 1.0 / len(newlist)
    return dic

def outputPriors4(ids, valuePairs1, valuePairs2, valuePairs3, valuePairs4, valuePairsRankRatio1, valuePairsRankRatio2, valuePairsRankRatio3, valuePairsRankRatio4, outputPath):
    idsList = sorted(list(ids))
    idValuePairsDic1 = {}
    for index in range(0, len(valuePairs1)):
        idValuePairsDic1[valuePairs1[index][0]] = valuePairs1[index][1]
    idValuePairsDic2 = {}
    for index in range(0, len(valuePairs2)):
        idValuePairsDic2[valuePairs2[index][0]] = valuePairs2[index][1]
    idValuePairsDic3 = {}
    for index in range(0, len(valuePairs3)):
        idValuePairsDic3[valuePairs3[index][0]] = valuePairs3[index][1]
    idValuePairsDic4 = {}
    for index in range(0, len(valuePairs4)):
        idValuePairsDic4[valuePairs4[index][0]] = valuePairs4[index][1]
        
    outputFile = open(outputPath, 'w')
    for id in idsList:
        if idValuePairsDic1.has_key(id):       
            rankRatio1 = valuePairsRankRatio1[idValuePairsDic1[id]]
            rankRatio2 = valuePairsRankRatio2[idValuePairsDic2[id]]
            rankRatio3 = valuePairsRankRatio3[idValuePairsDic3[id]]
            rankRatio4 = valuePairsRankRatio4[idValuePairsDic4[id]]
            value = sqrt((pow(rankRatio1, 2) + pow(rankRatio2, 2) + pow(rankRatio3, 2) + pow(rankRatio4, 2)) / 4)
            outputFile.write(str(id)+"\t"+str(value) + "\n")
        else:
            outputFile.write(str(id)+"\t"+"0\n")
    outputFile.close()
   
def outputRole(Ids, PRKL, PREntro, DegKL, DegEntro, resultBase):
    PRKLRankRatioDic = findRankRatio([a[1] for a in PRKL], rev=True)
    PREntroRankRatioDic = findRankRatio([a[1] for a in PREntro])
    DegKLRankRatioDic = findRankRatio([a[1] for a in DegKL], rev=True)
    DegEntroRankRatioDic = findRankRatio([a[1] for a in DegEntro])
    outputPriors4(Ids, PRKL, PREntro, DegKL, DegEntro, PRKLRankRatioDic, PREntroRankRatioDic, DegKLRankRatioDic, DegEntroRankRatioDic, resultBase)

def getUserPR(userIds,nodesPR):
    userPR = []
    for user in userIds:
        userPR.append(nodesPR[user])
    return userPR

def getUserDisctri(userAdj):
    userDegDistri = []
    for user in userAdj:
        userDegDistri.append(len(userAdj[user]))
    return userDegDistri

def extractFeatures(configure):
    nodeCount, nodesPR = readPR(configure.PRPath)
    prodAdj, userAdj,productIDs = getAdj(nodesPR, nodeCount, configure.edgeListPath)
    userPR = getUserPR(userAdj.keys(),nodesPR)
    neighUserPRDistri, neighUserDegDistri = findHighDegProd(prodAdj, userAdj, nodesPR, configure.highDegThresh)
    userDegDistri = getUserDisctri(userAdj)
         
    globalPRDistri, bins = vector2ProbHistPR(userPR, configure)  # compare with neighbors, so it is "userPR"
    prodPRKL = []
    prodPREntro = []
    for prodEntry in neighUserPRDistri:
        Id, KL = computeKL(prodEntry, globalPRDistri, bins, configure.epslon)
        Id, Entro = computeEntro(prodEntry, bins)
        prodPRKL.append([Id, KL])
        prodPREntro.append([Id, Entro])
    
    globalDegDistri, bins = vector2ProbHistDeg(userDegDistri, configure)
    prodDegKL = []
    prodDegEntro = []
    for prodEntry in neighUserDegDistri:
        Id, KL = computeKL(prodEntry, globalDegDistri, bins, configure.epslon)
        Id, Entro = computeEntro(prodEntry, bins)
        prodDegKL.append([Id, KL])
        prodDegEntro.append([Id, Entro])
    
    return productIDs, prodPRKL, prodPREntro, prodDegKL, prodDegEntro
    
if __name__ == '__main__':
    inputfileDir = sys.argv[1] 
    prfileDir = sys.argv[2] 
    outputDir= sys.argv[3]  
    figDir= sys.argv[4] 
    configure = Config(inputfileDir,prfileDir,outputDir,figDir)
    print "======Computing NSF values..."
    productIDs, prodPRKL, prodPREntro, prodDegKL, prodDegEntro = extractFeatures(configure)
    print "===========Step1: compute all feature values, done"
    plotFeatures(configure, prodPRKL, prodPREntro, prodDegKL, prodDegEntro)
    print "===========Step2: plot Features, done"
    outputRole(productIDs, prodPRKL, prodPREntro, prodDegKL, prodDegEntro, configure.priorPath)
    print "===========Step3: output NFS values, done"
