function []=computePR(fname,outputFile,epsilon)
% function: compute pagerank values for undirected graphs
% fname: input file name with format "nodeId1,nodeId2" for each line , nodeId is consective from 0 to n-1
% outputFile: output file store each node's pagerank value, like "0.00123" in each line

disp('======Computing Pagerank values...')
data = textread(strcat('',fname),'','delimiter',',');

nodesNum = max(max(data(:,1)), max(data(:,2)))+1;

G = sparse([data(:,1);data(:,2)]+1,[data(:,2);data(:,1)]+1,1,nodesNum,nodesNum);
p = 0.85;

c = sum(G,1);
k = find(c~=0);

z = ((1-p)*(c~=0) + (c==0))/nodesNum;

D = sparse(k,k,1./c(k),nodesNum,nodesNum);
e = ones(nodesNum,1);
G = G*D;

G = p*G;

x = e/nodesNum;
oldx = zeros(nodesNum,1);
diffV = 1;
while diffV > epsilon
   oldx = x;
   x = G*x + e*(z*x);
   diffV = norm(x-oldx);
end
x = x/sum(x);

fileID = fopen(strcat(outputFile),'w');
fprintf(fileID,'%.15f\n',x);
fclose(fileID);
disp('======Done!')
exit;
